import collections
import enum
import random
import string
import typing as t


__all__ = [
    "CircularChar",
    "CircularString",
    "ColorCode",
    "SerialString",
    "insert_colorcode",
    "rand_string",
]


class CircularChar:

    # NOTE
    #   This order is compatible with database.
    ROW = string.digits + string.ascii_uppercase + string.ascii_lowercase
    END = len(ROW) - 1

    HEAD = ROW[0]
    TAIL = ROW[-1]

    def __init__(self, begin: str = "0") -> None:
        if len(begin) != 1:
            raise ValueError("'begin' must be one charactor.")
        if begin not in self.ROW:
            raise ValueError("'begin' must be an ASCII charactor or digit.")

        self._index = 0

        for i, char in enumerate(self.ROW):
            if begin == char:
                self._index = i
                break

    @property
    def current(self) -> str:
        return self.ROW[self._index]

    def next(self, up: bool = True) -> str:
        if self._index == self.END:
            index = 0
        else:
            index = self._index + 1

        if up:
            self._index = index

        return self.ROW[index]

    def back(self, down: bool = True) -> str:
        if self._index == 0:
            index = self.END
        else:
            index = self._index - 1

        if down:
            self._index = index

        return self.ROW[index]


class SerialString:

    def __init__(self, length: int, begin: t.Optional[str] = None) -> None:
        if length < 1:
            raise ValueError("'length' must be bigger than 0.")

        if begin and len(begin) != length:
            raise ValueError(
                f"The length of 'begin' {begin} doesn't match {length}."
            )

        self._is_full = False
        self._is_first = False
        if begin:
            self._string = [CircularChar(b) for b in begin]
        else:
            self._string = [CircularChar() for _ in range(length)]
            self._is_first = True

    @property
    def current(self) -> str:
        return "".join([c.current for c in self._string])

    def next(self, up: bool = True) -> t.Optional[str]:
        result = collections.deque()

        def challenge(string: t.List[CircularChar], is_carry_up: bool) -> None:
            next_up = False
            if is_carry_up:
                char = string[-1].next(up=up)
                if char == CircularChar.HEAD:
                    next_up = True
            else:
                char = string[-1].current

            result.appendleft(char)

            if len(string) == 1:
                if next_up:
                    self._is_full = True
                return

            challenge(string[:-1], next_up)

        if self._is_first:
            challenge(self._string, False)
            self._is_first = False
        else:
            challenge(self._string, True)

        if self._is_full:
            return None

        return "".join(result)


class CircularString:

    def __init__(self, length: int, begin: t.Optional[str] = None) -> None:
        if length < 1:
            raise ValueError("'length' must be bigger than 0.")

        if begin and len(begin) != length:
            raise ValueError(
                f"The length of 'begin' {begin} doesn't match {length}."
            )

        if begin:
            self._string = [CircularChar(b) for b in begin]
        else:
            self._string = [CircularChar() for _ in range(length)]

    @property
    def current(self) -> str:
        return "".join([c.current for c in self._string])

    def next(self, up: bool = True) -> str:
        result = collections.deque()

        def challenge(string: t.List[CircularChar], is_carry_up: bool) -> None:
            next_up = False
            if is_carry_up:
                char = string[-1].next(up=up)
                if char == CircularChar.HEAD:
                    next_up = True
            else:
                char = string[-1].current

            result.appendleft(char)
            if len(string) == 1:
                return

            challenge(string[:-1], next_up)

        challenge(self._string, True)
        return "".join(result)

    def back(self, down: bool = True) -> str:
        result = collections.deque()

        def challenge(
            string: t.List[CircularChar],
            is_carry_down: bool
        ) -> None:
            next_down = False
            if is_carry_down:
                char = string[-1].back(down=down)
                if char == CircularChar.TAIL:
                    next_down = True
            else:
                char = string[-1].current

            result.appendleft(char)

            if len(string) == 1:
                return

            challenge(string[:-1], next_down)

        challenge(self._string, True)
        return "".join(result)


_ASCII_LETTERS_DIGITS = string.ascii_letters + string.digits


def rand_string(size: int) -> str:
    return "".join(random.choices(_ASCII_LETTERS_DIGITS, k=size))


class ColorCode(enum.Enum):

    BLACK       = "\033[30m"
    RED         = "\033[31m"
    GREEN       = "\033[32m"
    YELLOW      = "\033[33m"
    BLUE        = "\033[34m"
    MAGENTA     = "\033[35m"
    CYAN        = "\033[36m"
    WHITE       = "\033[37m"
    DEFAULT     = "\033[39m"
    BOLD        = "\033[1m"
    UNDERLINE   = "\033[4m"
    INVISIBLE   = "\033[08m"
    REVERCE     = "\033[07m"

    BG_BLACK    = "\033[40m"
    BG_RED      = "\033[41m"
    BG_GREEN    = "\033[42m"
    BG_YELLOW   = "\033[43m"
    BG_BLUE     = "\033[44m"
    BG_MAGENTA  = "\033[45m"
    BG_CYAN     = "\033[46m"
    BG_WHITE    = "\033[47m"
    BG_DEFAULT  = "\033[49m"

    RESET       = "\033[0m"


def insert_colorcode(text: str, code: ColorCode, reset: bool = True) -> str:
    text = "".join((code.value, text))
    if reset:
        text = "".join((text, ColorCode.RESET.value))

    return text



import random
import string
import typing as t
import unittest

from bamboo import (
    AnyStringLocation,
    AsciiDigitLocation,
    DuplicatedUriRegisteredError,
    Router,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http

from . import get_log_name


IDEAL_RESNPONSE = "Hello, World".encode()
PATH_SERVER_LOG = get_log_name(__file__)


def rand_string(k: int) -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=k))


def rand_strings(num: int) -> t.List[str]:
    return [rand_string(random.randint(10, 20)) for _ in range(num)]


# Template endpoints
class MockEndpoint(WSGIEndpoint):

    def do_GET(self) -> None:
        self.send_body(IDEAL_RESNPONSE)


class TestRouting(unittest.TestCase):

    def test_routing(self):
        # Make different uris
        paths = list(set(rand_strings(30)))

        def client():
            for i, path in enumerate(paths):
                uri = f"http://localhost:{8000 + i}/{path}"
                with http.get(uri) as res:
                    self.assertEqual(res.body, IDEAL_RESNPONSE)

        forms = []
        for i, path in enumerate(paths):
            app = WSGIApp()
            app.route(path)(MockEndpoint)
            forms.append(WSGIServerForm("", 8000 + i, app, PATH_SERVER_LOG))

        WSGITestExecutor(*forms).exec(client)

    def assertDuplicatedUris(self, patterns):
        router = Router()

        with self.assertRaises(DuplicatedUriRegisteredError) as err:
            for uri in patterns:
                router.register(uri, MockEndpoint)

        self.assertIsInstance(err.exception, DuplicatedUriRegisteredError)

    def test_duplicated_uris(self):
        pattern_1 = [
            ("test", "hoge", AnyStringLocation()),
            ("test", "hoge", "image")
        ]
        pattern_2 = [
            (AnyStringLocation(), AsciiDigitLocation(4)),
            ("hoge", "hoge")
        ]
        pattern_3 = [
            (AnyStringLocation(), "test", AnyStringLocation()),
            ("hoge", AnyStringLocation(), "image")
        ]

        self.assertDuplicatedUris(pattern_1)
        self.assertDuplicatedUris(pattern_2)
        self.assertDuplicatedUris(pattern_3)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import dataclasses
import json as js
import typing as t
from urllib.parse import parse_qs, urlparse

from ..api import JsonApiData
from ..base import HTTPMethods, MediaTypes
from ..util.convert import unparse_qs


__all__ = []


@dataclasses.dataclass
class HTTPRequestForm:

    host: str
    port: t.Optional[int]
    uri: str
    method: str
    headers: t.Dict[str, str]
    body: t.Optional[bytes]


def get_http_request_form(
    scheme: str,
    uri: str,
    method: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData, None] = None,
    query: t.Dict[str, t.List[str]] = {}
) -> HTTPRequestForm:
    # method management
    method = method.upper()
    if method not in HTTPMethods:
        raise ValueError(f"Specified method '{method}' is not available.")

    # body management
    if body and json:
        raise ValueError("Request body is specified both 'body' and 'json'.")
    if body:
        if "Content-Type" not in headers:
            headers["Content-Type"] = MediaTypes.plain
    if json:
        if isinstance(json, JsonApiData):
            json = json.dict
        body = js.dumps(json)
        if "Content-Type" not in headers:
            headers["Content-Type"] = MediaTypes.json

    parsed_uri = urlparse(uri)
    if parsed_uri.scheme != scheme:
        raise ValueError(
            f"Scheme of specified uri '{parsed_uri.scheme}' is "
            "not available. Use HTTP."
        )

    # port management
    port = parsed_uri.port
    if not port:
        port = None

    # query management
    query_included = parse_qs(parsed_uri.query)
    query_included.update(query)
    query = unparse_qs(query_included)

    # path management
    path = parsed_uri.path
    if len(query):
        path = "?".join((path, query))

    return HTTPRequestForm(
        parsed_uri.hostname,
        port,
        path,
        method,
        headers,
        body
    )



import os
from pathlib import Path


NAME_IMAGE = "elephant.jpg"
NAME_BIGDATA_1KB = "bigdata_1kb"
NAME_BIGDATA_1MB = "bigdata_1mb"
NAME_BIGDATA_1GB = "bigdata_1gb"

DIR_TESTS = str(Path(__file__).parent)
DIR_RES = os.path.join(DIR_TESTS, "res")
PATH_IMAGE = os.path.join(DIR_RES, NAME_IMAGE)
PATH_BIGDATA_1KB = os.path.join(DIR_RES, NAME_BIGDATA_1KB)
PATH_BIGDATA_1MB = os.path.join(DIR_RES, NAME_BIGDATA_1MB)
PATH_BIGDATA_1GB = os.path.join(DIR_RES, NAME_BIGDATA_1GB)


def make_bigdata() -> None:
    if not os.path.isfile(PATH_BIGDATA_1KB):
        with open(PATH_BIGDATA_1KB, "wb") as f_1k:
            f_1k.write(os.getrandom(10**3))

    if not os.path.isfile(PATH_BIGDATA_1MB):
        with open(PATH_BIGDATA_1MB, "wb") as f_1m:
            with open(PATH_BIGDATA_1KB, "rb") as f_1k:
                data_1k = f_1k.read()

            for _ in range(10**3):
                f_1m.write(data_1k)

    if not os.path.isfile(PATH_BIGDATA_1GB):
        with open(PATH_BIGDATA_1GB, "wb") as f_1g:
            with open(PATH_BIGDATA_1MB, "rb") as f_1m:
                data_1m = f_1m.read()

            for _ in range(10**3):
                f_1g.write(data_1m)


def get_fname(fpath: str) -> str:
    return fpath.split(".")[0]


def get_log_name(fpath: str, *suffix: str) -> str:
    fname = [get_fname(fpath)]
    fname.extend(suffix)
    return f"{'_'.join(fname)}.log"


def get_client_log_name(fpath: str, *suffix: str) -> str:
    fname = [get_fname(fpath)]
    fname.extend(suffix)
    return f"{'_'.join(fname)}_client.log"


make_bigdata()



import sys
import typing as t


if sys.version_info.minor >= 8:
    from typing import get_args, get_origin
else:
    # NOTE
    #   Below is for compatability to Python 3.7.
    #   The implementation is same as the one of Python 3.8.

    import collections

    def get_origin(tp):
        if isinstance(tp, t._GenericAlias):
            return tp.__origin__
        if tp is t.Generic:
            return t.Generic
        return None

    def get_args(tp):
        if isinstance(tp, t._GenericAlias) and not tp._special:
            res = tp.__args__
            if (
                get_origin(tp) is collections.abc.Callable and
                res[0] is not Ellipsis
            ):
                res = (list(res[:-1]), res[-1])
            return res
        return ()



import random
import typing as t
import unittest

from bamboo import (
    ASGIHTTPApp,
    ASGIHTTPEndpoint,
    JsonApiData,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http
from bamboo.sticky.http import data_format
from bamboo.util.string import rand_string

from . import get_log_name
from .asgi_util import ASGIHTTPServerForm, ASGIHTTPTestExecutor


app_asgi = ASGIHTTPApp()
app_wsgi = WSGIApp()
PATH_ASGI_SERVER_LOG = get_log_name(__file__, "asgi")
PATH_WSGI_SERVER_LOG = get_log_name(__file__, "wsgi")


class UserJsonApi(JsonApiData):

    user_id: str
    name: str
    email: str
    age: int


class TestJsonApi(JsonApiData):

    users: t.List[UserJsonApi]
    total: int


def make_user_api_data() -> t.Dict[str, t.Any]:
    user_id = rand_string(10)
    name = rand_string(10)
    email = rand_string(20)
    age = random.randint(0, 100)
    return {
        "user_id": user_id,
        "name": name,
        "email": email,
        "age": age,
    }


def make_test_api_data() -> t.Dict[str, t.Any]:
    total = random.randint(0, 100)
    users = [make_user_api_data() for _ in range(total)]
    return {
        "users": users,
        "total": total,
    }


@app_asgi.route()
class TestASGIHTTPEndpoint(ASGIHTTPEndpoint):

    @data_format(input=None, output=TestJsonApi)
    async def do_GET(self) -> None:
        assert await self.body == b""
        self.send_json(make_test_api_data())

    @data_format(input=TestJsonApi, output=None)
    async def do_POST(self, rec_body: TestJsonApi) -> None:
        self.send_only_status()

    @data_format(input=JsonApiData, output=None, is_validate=False)
    async def do_DELETE(self) -> None:
        self.send_only_status()


@app_wsgi.route()
class TestWSGIEndpoint(WSGIEndpoint):

    @data_format(input=None, output=TestJsonApi)
    def do_GET(self) -> None:
        assert self.body == b""
        self.send_json(make_test_api_data())

    @data_format(input=TestJsonApi, output=None)
    def do_POST(self, rec_body: TestJsonApi) -> None:
        self.send_only_status()

    @data_format(input=JsonApiData, output=None, is_validate=False)
    def do_DELETE(self) -> None:
        self.send_only_status()


class TestStickyDataFormat(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form_asgi = ASGIHTTPServerForm("", 8000, app_asgi, PATH_ASGI_SERVER_LOG)
        form_wsgi = WSGIServerForm("", 8001, app_wsgi, PATH_WSGI_SERVER_LOG)
        cls.executor_asgi = ASGIHTTPTestExecutor(form_asgi).start_serve()
        cls.executor_wsgi = WSGITestExecutor(form_wsgi).start_serve()
        cls.uri_asgi = "http://localhost:8000"
        cls.uri_wsgi = "http://localhost:8001"

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor_asgi.close()
        cls.executor_wsgi.close()

    def test_asgi_output(self):
        with http.get(self.uri_asgi) as res:
            self.assertTrue(res.ok)
            api_data = res.attach(TestJsonApi)
            self.assertIsInstance(api_data, TestJsonApi)

    def test_asgi_input_validation(self):
        data = make_test_api_data()
        with http.post(self.uri_asgi, json=data) as res:
            self.assertTrue(res.ok)

    def test_asgi_input_no_validation(self):
        data = make_test_api_data()
        with http.delete(self.uri_asgi, json=data) as res:
            self.assertTrue(res.ok)

    def test_wsgi_output(self):
        with http.get(self.uri_wsgi) as res:
            self.assertTrue(res.ok)
            api_data = res.attach(TestJsonApi)
            self.assertIsInstance(api_data, TestJsonApi)

    def test_wsgi_input_validation(self):
        data = make_test_api_data()
        with http.post(self.uri_wsgi, json=data) as res:
            self.assertTrue(res.ok)

    def test_wsgi_input_no_validation(self):
        data = make_test_api_data()
        with http.delete(self.uri_wsgi, json=data) as res:
            self.assertTrue(res.ok)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import http.client
import typing as t

from ..api import BinaryApiData, JsonApiData
from ..base import HTTPMethods
from ..request import ResponseData_t, Schemes
from ..request.request_form import get_http_request_form
from ..request.response import Response


__all__ = [
    "connect",
    "delete",
    "get",
    "head",
    "options",
    "patch",
    "post",
    "put",
    "trace",
]


def request(
    uri: str,
    method: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    form = get_http_request_form(
        Schemes.HTTP,
        uri,
        method,
        headers=headers,
        body=body,
        json=json,
        query=query
    )
    conn = http.client.HTTPConnection(
        form.host,
        port=form.port,
        timeout=timeout,
        blocksize=blocksize
    )
    conn.request(form.method, form.uri, body=form.body, headers=form.headers)
    _res = conn.getresponse()
    return Response(conn, _res, datacls=datacls)


def get(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the GET method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.GET,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )


def post(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the POST method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.POST,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )


def put(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the PUT method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.PUT,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )


def delete(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the DELETE method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.DELETE,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )


def head(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the HEAD method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.HEAD,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )


def options(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the OPTIONS method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.OPTIONS,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )


def patch(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the PATCH method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.PATCH,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )


def trace(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the TRACE method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.TRACE,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )


def connect(
    uri: str,
    headers: t.Dict[str, str] = {},
    body: t.Optional[bytes] = None,
    json: t.Union[t.Dict[str, t.Any], JsonApiData] = None,
    query: t.Dict[str, t.List[str]] = {},
    timeout: t.Optional[float] = None,
    blocksize: int = 8192,
    datacls: t.Type[ResponseData_t] = BinaryApiData,
) -> Response[ResponseData_t]:
    """Request with the CONNECT method on HTTP.

    Args:
        uri: URI to be requested.
        headers: Request headers.
        body: Request body of bytes.
        json: Request body of JSON.
        query: Query parameters to be attached to the URI.
        timeout: Seconds waiting for the connection.
        blocksize: Block size of sending data.
        datacls: `ApiData` or its subclass to be attached from the response body.

    Returns:
        Response object generated with the response.
    """
    return request(
        uri,
        HTTPMethods.CONNECT,
        headers=headers,
        body=body,
        json=json,
        query=query,
        timeout=timeout,
        blocksize=blocksize,
        datacls=datacls
    )



import typing as t

from bamboo.endpoint import ASGIHTTPEndpoint, WSGIEndpoint


__all__ = []


# Signature of the callback of response method on sub classes of
# the Endpoint class.
Callback_WSGI_t = t.Callable[[WSGIEndpoint, t.Tuple[t.Any, ...]], None]
Callback_ASGI_t = t.Callable[[ASGIHTTPEndpoint, t.Tuple[t.Any, ...]], t.Awaitable[None]]
Callback_t = t.Union[Callback_WSGI_t, Callback_ASGI_t]
CallbackDecorator_t = t.Callable[[Callback_t], Callback_t]


def _get_bamboo_attr(attr: str) -> str:
    return f"__bamboo_{attr}__"


class DuplicatedInfoError(Exception):
    """Raised if same type of infomation is duplicated."""
    pass



import os
import time
import unittest

from bamboo import (
    BinaryApiData,
    ASGIHTTPApp,
    ASGIHTTPEndpoint,
)
from bamboo.request import http
from bamboo.sticky.http import data_format

from . import (
    NAME_BIGDATA_1GB,
    NAME_BIGDATA_1KB,
    NAME_BIGDATA_1MB,
    PATH_BIGDATA_1GB,
    PATH_BIGDATA_1KB,
    PATH_BIGDATA_1MB,
    get_log_name,
    get_client_log_name
)
from .asgi_util import ASGIHTTPServerForm, ASGIHTTPTestExecutor


app = ASGIHTTPApp()
PATH_SERVER_LOG = get_log_name(__file__)
PATH_CLIENT_LOG = get_client_log_name(__file__)
FILE_CLIENT_LOG = open(PATH_CLIENT_LOG, "wt")


@app.route("small")
class SmallFileServingEndpoint(ASGIHTTPEndpoint):

    @data_format(input=None, output=BinaryApiData)
    async def do_GET(self) -> None:
        self.send_file(PATH_BIGDATA_1KB, NAME_BIGDATA_1KB)


@app.route("medium")
class MediumFileServingEndpoint(ASGIHTTPEndpoint):

    @data_format(input=None, output=BinaryApiData)
    async def do_GET(self) -> None:
        self.send_file(PATH_BIGDATA_1MB, NAME_BIGDATA_1MB)


@app.route("big")
class BigFileServingEndpoint(ASGIHTTPEndpoint):

    @data_format(input=None, output=BinaryApiData)
    async def do_GET(self) -> None:
        self.send_file(PATH_BIGDATA_1GB, NAME_BIGDATA_1GB)


class WSGISendingFileTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form = ASGIHTTPServerForm("", 8000, app, PATH_SERVER_LOG)
        cls.executor = ASGIHTTPTestExecutor(form).start_serve()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor.close()
        FILE_CLIENT_LOG.close()

    def measure_downloading(self, uri: str, fpath: str) -> None:
        filesize = os.path.getsize(fpath)
        time_init = time.time()
        bufsize = 8192

        with http.get(uri) as res:
            with open(fpath, "rb") as fsrc:
                total = 0
                while True:
                    chunk = res.read(bufsize)
                    chunksize = len(chunk)
                    if not chunk:
                        break

                    total += chunksize
                    chunk_src = fsrc.read(chunksize)
                    self.assertEqual(chunk, chunk_src)
                self.assertEqual(total, filesize)

        time_total = time.time() - time_init
        FILE_CLIENT_LOG.write(
            f"Time of downloading {fpath} : {time_total} sec\n"
        )

    def test_small(self) -> None:
        uri = "http://localhost:8000/small"
        self.measure_downloading(uri, PATH_BIGDATA_1KB)

    def test_medium(self) -> None:
        uri = "http://localhost:8000/medium"
        self.measure_downloading(uri, PATH_BIGDATA_1MB)

    def test_big(self) -> None:
        uri = "http://localhost:8000/big"
        self.measure_downloading(uri, PATH_BIGDATA_1GB)


if __name__ == "__main__":
    unittest.main()



import os
import typing as t
import unittest

from bamboo import BufferedConcatIterator, BufferedFileIterator

from . import PATH_IMAGE


def test_generator() -> t.Generator[bytes, None, None]:
    for i in range(10**2, 10**3):
        yield os.getrandom(i)


class BufferedConcatIteratorTest(unittest.TestCase):

    def setUp(self) -> None:
        self.total = sum((
            10**5,
            sum(range(10**2, 10**3)),
            10**5,
            sum(range(10**2, 10**3)),
            os.path.getsize(PATH_IMAGE))
        )
        self.binary = os.getrandom(10**5)
        self.bufsize = 8192
        self.iter = BufferedConcatIterator(
            self.binary,
            test_generator(),
            self.binary,
            test_generator(),
            BufferedFileIterator(PATH_IMAGE),
            bufsize=self.bufsize
        )

    def test_iter(self):
        sum = 0
        for i in self.iter:
            diff = self.total - sum
            if diff < self.bufsize:
                self.assertEqual(len(i), diff)
            else:
                self.assertEqual(len(i), self.bufsize)

            sum += len(i)
        self.assertEqual(sum, self.total)


if __name__ == "__main__":
    unittest.main()



import json
import typing as t
import unittest

from bamboo import (
    ContentType,
    JsonApiData,
    MediaTypes,
    ApiValidationFailedError,
)


json_content_type = ContentType(MediaTypes.json, "utf-8")
not_json_content_type = ContentType(MediaTypes.plain, "utf-8")


class TestInnerData(JsonApiData):

    name: str
    age: int
    email: str


data_inner = json.dumps({
    "age": 18, "email": "hoge@hoge.com", "name": "hogehoge"
}).encode()


class TestListData(JsonApiData):

    accounts: t.List[TestInnerData]
    datetime: str


data_list = json.dumps({
    "accounts": [
        {"name": "hogehoge", "age": 18, "email": "hoge@hoge.com"},
        {"name": "hogehoge", "age": 18, "email": "hoge@hoge.com"},
        {"name": "hogehoge", "age": 18, "email": "hoge@hoge.com"},
        {"name": "hogehoge", "age": 18, "email": "hoge@hoge.com"},
        {"name": "hogehoge", "age": 18, "email": "hoge@hoge.com"},
        {"name": "hogehoge", "age": 18, "email": "hoge@hoge.com"},
        {"name": "hogehoge", "age": 18, "email": "hoge@hoge.com"},
    ],
    "datetime": "2000.01.01-00:00:00"
}).encode()


class TestNestedData(JsonApiData):

    account: TestInnerData
    datetime: str


data_nested = json.dumps({
    "account": {"name": "hogehoge", "age": 18, "email": "hoge@hoge.com"},
    "datetime": "2000.01.01-00:00:00"
}).encode()


class TestUnionData(JsonApiData):

    name: str
    age: t.Optional[int] = None


data_union = json.dumps({
    "name": "hogehoge", "age": None,
}).encode()

data_invalid_type = json.dumps({
    "name": 19, "age": 20,
}).encode()

data_key_not_included = json.dumps({
    "Name": "hogehoge", "age": 30,
}).encode()

data_default_value = json.dumps({
    "name": "hogehoge"
}).encode()


class TestJsonApiData(unittest.TestCase):

    def test_inner_data(self):
        data = TestInnerData.__validate__(data_inner, json_content_type)
        self.assertEqual(data.name, "hogehoge")
        self.assertEqual(data.age, 18)
        self.assertEqual(data.email, "hoge@hoge.com")

    def test_list_data(self):
        data = TestListData.__validate__(data_list, json_content_type)
        self.assertTrue(isinstance(data.accounts, list))
        self.assertTrue(isinstance(data.accounts[0], TestInnerData))
        self.assertEqual(data.accounts[0].name, "hogehoge")
        self.assertEqual(data.accounts[0].age, 18)
        self.assertEqual(data.accounts[0].email, "hoge@hoge.com")
        self.assertEqual(data.datetime, "2000.01.01-00:00:00")

    def test_nested_data(self):
        data = TestNestedData.__validate__(data_nested, json_content_type)
        self.assertTrue(isinstance(data.account, TestInnerData))
        self.assertEqual(data.account.name, "hogehoge")
        self.assertEqual(data.account.age, 18)
        self.assertEqual(data.account.email, "hoge@hoge.com")
        self.assertEqual(data.datetime, "2000.01.01-00:00:00")

    def test_union_data(self):
        data = TestUnionData.__validate__(data_union, json_content_type)
        self.assertEqual(data.name, "hogehoge")
        self.assertEqual(data.age, None)

    def test_invalid_type(self):
        with self.assertRaises(ApiValidationFailedError) as err:
            data = TestUnionData.__validate__(data_invalid_type, json_content_type)
        self.assertIsInstance(err.exception, ApiValidationFailedError)

    def test_key_not_included(self):
        with self.assertRaises(ApiValidationFailedError) as err:
            data = TestUnionData.__validate__(data_key_not_included, json_content_type)
        self.assertIsInstance(err.exception, ApiValidationFailedError)

    def test_not_json_content_tyep(self):
        with self.assertRaises(ApiValidationFailedError) as err:
            data = TestUnionData.__validate__(data_union, not_json_content_type)
        self.assertIsInstance(err.exception, ApiValidationFailedError)

    def test_default_value(self):
        data = TestUnionData.__validate__(data_default_value, json_content_type)
        self.assertEqual(data.name, "hogehoge")
        self.assertEqual(data.age, None)


if __name__ == "__main__":
    unittest.main()



import typing as t
import unittest

from bamboo.api import (
    InvalidAnnotationError,
    JsonApiData,
    JsonApiDataBuilder,
)


class TestInnerApi(JsonApiData):

    name: str
    age: int


class TestOuterApi(JsonApiData):

    accounts: t.List[TestInnerApi]


class TestJsonable(unittest.TestCase):

    def test_int(self):
        JsonApiDataBuilder.check_annotations(int)

    def test_float(self):
        JsonApiDataBuilder.check_annotations(float)

    def test_str(self):
        JsonApiDataBuilder.check_annotations(str)

    def test_bool(self):
        JsonApiDataBuilder.check_annotations(bool)

    def test_NoneType(self):
        JsonApiDataBuilder.check_annotations(type(None))

    def test_list(self):
        JsonApiDataBuilder.check_annotations(t.List[int])

        with self.assertRaises(InvalidAnnotationError) as err:
            JsonApiDataBuilder.check_annotations(t.List)
        self.assertIsInstance(err.exception, InvalidAnnotationError)

    def test_jsonapi(self):
        JsonApiDataBuilder.has_valid_annotations(TestOuterApi)


if __name__ == "__main__":
    unittest.main()
